@@ -695,6 +695,10 @@ class DPLCOMMONDriver(driver.CloneableImageVD,
                              cert_path=cert_path)
         self._stats = {}
 
+    @staticmethod
+    def get_driver_options():
+        return options.DPL_OPTS
+
     def _convert_size_GB(self, size):
         s = round(float(size) / units.Gi, 2)
         if s > 0:
@@ -1438,9 +1442,6 @@ class DPLCOMMONDriver(driver.CloneableImageVD,
                 pool['free_capacity_gb'] = (
                     self._convert_size_GB(
                         int(output['metadata']['available_capacity'])))
-                pool['allocated_capacity_gb'] = (
-                    self._convert_size_GB(
-                        int(output['metadata']['used_capacity'])))
                 pool['QoS_support'] = False
                 pool['reserved_percentage'] = 0
                 pools.append(pool)
